# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Test Result Analysis.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
# Andre Heinecke <andre.heinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
import datetime
import gzip
import json
import os
import Queue
import shutil
import tarfile
import time
import unittest

from sqlalchemy import *

from mapwoc.lib.hc import HC
from mapwoc.lib.hccontrol import HCControl
from mapwoc.lib.message import Message
from mapwoc.lib.redirectorvisit import RedirectorVisit
from mapwoc.lib.scanconfiguration import ScanConfiguration, VisitorBundle, VisitorConfiguration
from mapwoc.lib.trivialresponse import TrivialResponse
from mapwoc.lib.url import URL, create_URL_List
from mapwoc.lib.mapwocqueue import Mapwocqueue
from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook
from mapwoc.lib.model.result import *
from mapwoc.nodectrl.analyserprocess import AnalysisProcess
from mapwocgui.model.meta import Session
from mapwoc.masterctrl.resultfileanalyser import ResultFileAnalyser

from tests.util import enable_logging

# Script to be used for analysis
ANALYSERSCRIPT = "analyser/src/mapwocanalyse.pl"

# URLs that will be commited to the database
AVAILABLE_URLS = [ "http://www.intevation.de",
                   "http://www.sourceforge.net",
                   "http://www.jkw-soundcafe.de",
                   "http://www.ftd.de" ]

# TODO: check test verbosity if this is possible
enable_logging()

def visits_for_resulttype(result_name):
    """
    Query the database for visits that contain a specific resulttype.

    @param result_name Name of the resulttype
    @return visits that contain the result
    """

    result_type = Session.query(ResultType).filter_by(
                                type_name=result_name).first()

    visits = Session.query(Visit).filter(
                Visit.results.any(result_type_id = result_type.id)).all()

    return visits

class TestAnalyseBase(unittest.TestCase):

    def setUp(self):
        self.urllist, _invalids = create_URL_List("testurls",
                                        "\n".join(AVAILABLE_URLS))
        self.scanconfiguration = ScanConfiguration("testscan")
        self.scanconfiguration.urlcollection.append(self.urllist)
        self.scanconfiguration.comment = "Test Comment"
        self.scanconfiguration.description = "Test Description"
        self.scanconfiguration.visitor_configuration = VisitorConfiguration()
        self.scanrun = ScanRun(self.scanconfiguration, "Test")
        self.scanrun.stoptime = datetime.datetime.now()
        if not os.path.isdir("tests/tmp"):
            os.mkdir("tests/tmp")
        self.dbpath = "tests/tmp/tmpdb.db"
        shutil.copyfile("tests/dbs/database.db", self.dbpath)
        engine = create_engine('sqlite:///tests/tmp/tmpdb.db')
        __c = engine.connect()
        Session.configure(bind=engine)
        Session.add(self.urllist)
        for url in self.urllist.urls:
            Session.add(url)
        Session.add(self.scanconfiguration.visitor_configuration)
        Session.add(self.scanconfiguration)
        Session.add(self.scanrun)
        Session.flush()
        Session.commit()

    def tearDown(self):
        Session.remove() # Necessary to "clear the cache"
        os.remove("tests/tmp/tmpdb.db")

    def run_analyser(self, dumpfile):
        """
        Create a analyserprocess run it on one dumpfile, kill it.
        @return Path to result tarball.
        """
        job_queue = Mapwocqueue()
        result_queue = Mapwocqueue()
        work_dir = "tests/tmp/analyser_work"
        result_dir = "tests/tmp/analyser_result"

        for tdir in [work_dir, result_dir]:
            if not os.path.isdir(tdir):
                os.mkdir(tdir)

        analyser = AnalysisProcess(job_queue,
                                   result_queue,
                                   work_dir,
                                   result_dir,
                                   ANALYSERSCRIPT)
        analyser.start()
        job_queue.put(dumpfile)
        try:
            result = result_queue.get(True, 60)
        except Queue.Empty:
            self.assertEqual(True, False,
                        "AnalysisProcess did not produce a result "
                        "within 60 seconds for dump: %s" % dumpfile)
        finally:
            analyser.terminate()

        # Code duplication from node.pack_result because
        # of scanrun and so on are attributes from node
        tgz = tarfile.open(name=result, mode='a')
        metafile = os.path.join(work_dir, "metainfo")
        if os.path.isfile(metafile):
            logger.warning("Metafile %s already exists, deleting...")
            remove(metafile)
        with open(metafile, "w") as fptr:
            parser = ConfigParser.RawConfigParser()
            parser.add_section("scan")
            parser.set("scan", "time", time.time())
            parser.set("scan", "scanrunid", self.scanrun.id)
            parser.write(fptr)

        tgz.add(metafile, arcname="metainfo")
        tgz.close()
        f_in = open(result, 'rb')
        f_out = gzip.open(result + ".gz", 'wb')
        f_out.writelines(f_in)
        f_out.close()
        f_in.close()
        os.remove(result)
        result = result + ".gz"
        os.remove(metafile)
        return result

class TestComplicatedDump(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/winxp_ie6_favorit.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_is_fastmode(self):
        visits = Session.query(Visit).all()
        self.assertEqual(len(visits), 1, "There should be only one Visit per dump.")
        visit = visits[0]
        self.assertEqual(len(visit.urls), 2, "Not found the two urls of fastmode winxp_ie6 (complicated) dump.")

    def test_no_unknown_traffic(self):
        self.assertEqual(len(visits_for_resulttype("UnknownProtocol")), 0, "There should not be an UnknownProtocol result.")


class TestManualIntevationTraffic(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/winvista_ie7_base_manual_intevation.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_all(self):
        self.assertEqual(len(visits_for_resulttype("RegularTraffic")), 1)
        visits = Session.query(Visit).all()
        self.assertEqual(len(visits), 1, "There should be only one Visit per dump.")
        visit = visits[0]
        self.assertEqual(len(visit.urls), 0, "There should be no main url when manually browsed.")

class TestFastMode(TestAnalyseBase):

    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/win7_ie8_base_fastmode_files_kolab.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_all(self):
        self.assertEqual(len(visits_for_resulttype("RegularTraffic")), 1)
        visits = Session.query(Visit).all()
        self.assertEqual(len(visits), 1, "There should be five Visit per dump.")
        self.assertEqual(len(visits[0].urls), 5, "There should be five main urls in this visit.")

class TestResultFiles(TestAnalyseBase):
    # This is a single test because this behavior is expected to change in the future
    # when we will only save interesting results.
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/win7_ie8_base_manual_miscfiles.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_all(self):
        self.assertEqual(len(visits_for_resulttype("RegularTraffic")), 1)
        visits = Session.query(Visit).all()
        self.assertEqual(len(visits), 1, "There should be only one Visit per dump.")
        visit = visits[0]
        for result in visit.results:
            if result.result_type.type_name not in ["UnknownProtocol",
                    "404", "302", "204", "MissingTraffic", "GraphicsFound"]:
                self.assertEqual(not result.result_file, False, 
                        "Result: %s does not have a file" % \
                                result.result_type.type_name)


class TestRedirectedIntevationTraffic(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/winvista_ie7_base_redirect_intevation.dump",
          "tests/tmp/winvista_ie7_base-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/winvista_ie7_base-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/winvista_ie7_base-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_all(self):
        self.assertEqual(len(visits_for_resulttype("RegularTraffic")), 1)
        visits = Session.query(Visit).all()
        self.assertEqual(len(visits), 1, "There should be only one Visit per dump.")
        visit = visits[0]
        self.assertEqual(visit.hcsid,"winvista_ie7_base", "HCSID is: %s and not winvista_ie7_base" % \
                visit.hcsid)
        self.assertEqual(len(visit.urls), 1, "There should be a main url when manually browsed.")
        self.assertEqual(visit.urls[0].url, "http://www.intevation.de",
                "Main URL should be http://www.intevation.de but was: %s" % visit.urls[0].url )
        self.assertEqual(len(visits_for_resulttype("GoogleAnalyticsParser")), 0)
        self.assertEqual(len(visits_for_resulttype("404")), 0)
        self.assertEqual(len(visits_for_resulttype("WindowsExecutable")), 0)
        self.assertEqual(len(visits_for_resulttype("MissingTraffic")), 0)
        for result in visit.results:
            if result.url == "10.0.0.1":
                self.assertEqual(True, False,
                        "Redirector queries should be no results.")

class TestRegularTraffic(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/winvista_ie7_base_manual_intevation.dump",
          "tests/tmp/the_test_hcsid_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_all(self):
        self.assertEqual(len(visits_for_resulttype("RegularTraffic")), 1)
        visit = visits_for_resulttype("RegularTraffic")[0]
        self.assertEqual(visit.hcsid, "the_test_hcsid")


class TestMissingTraffic(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/no_traffic.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_all(self):
        visits = visits_for_resulttype("MissingTraffic")
        self.assertEqual(len(visits), 1)
        visit = visits[0]
        self.assertEqual(visit.hcsid,"the_test_hcsid", "HCSID is: %s and not the_test_hcsid" % \
                visit.hcsid)

class TestMimeTypeAnalysis(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/win7_ie8_base_manual_miscfiles.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self) :
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_Entry(self):
        self.assertEqual(len(visits_for_resulttype("MSOfficeDocument")), 1)
        self.assertEqual(len(visits_for_resulttype("MACExecutable")), 1)
        self.assertEqual(len(visits_for_resulttype("Flash")), 1)
        self.assertEqual(len(visits_for_resulttype("PDF")), 1)

class TestGoogleAnalytics(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/win7_ie8_base_manual_sourceforge.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_Entry(self):
        self.assertEqual(len(visits_for_resulttype("GoogleAnalyticsParser")), 1)
        self.assertEqual(len(visits_for_resulttype("GraphicsFound")), 1)

class TestWindowsExecutable(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/win7_ie8_base_manual_exe.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_Entry(self):
        self.assertEqual(len(visits_for_resulttype("WindowsExecutable")), 1)

    def test_sha1sum(self):
        visit = Session.query(Visit).first()
        for result in visit.results:
            if result.result_type.type_name == "WindowsExecutable":
                self.assertEqual("35239a56ecb4aa4d433ba98fbcd0a7ca8751913f",
                        result.result_file.sha1sum)

    def test_FileName(self):
        visit = Session.query(Visit).first()
        for result in visit.results:
            if result.result_type.type_name == "WindowsExecutable":
                self.assertEqual("widgettest.exe", result.text)

class TestVirusFound(TestAnalyseBase):
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/win7_ie8_base_manual_test_file.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_Entry(self):
        self.assertEqual(len(visits_for_resulttype("VirusFound")), 1)

class Test404(TestAnalyseBase):
    # It is not a error that manual_exe is used this dump also contains
    # 404's
    def setUp(self):
        shutil.copyfile(
          "tests/trafficdumps/win7_ie8_base_manual_exe.dump",
          "tests/tmp/the_test_hcsid-snapshot_123456.dump")
        TestAnalyseBase.setUp(self)
        result_tarball = self.run_analyser("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        r = ResultFileAnalyser(result_tarball, "nonode")
        r.analyse()

    def tearDown(self):
        try:
            os.remove("tests/tmp/the_test_hcsid-snapshot_123456.dump")
        except:
            pass
        TestAnalyseBase.tearDown(self)

    def test_Entry(self):
        self.assertEqual(len(visits_for_resulttype("404")), 1,
         "404 Result should be present in analysis of tests/trafficdumps/win7_ie8_base_manual_exe.dump")
